/*
  Copyright 2000, Charles L. Blake.  All rights reserved.  No warranty.  Etc.
  See the man page for details on the intentions of this code.
  NOTE: Pains were taken to use few libc functions to achieve tiny statically
        linked binaries. In particular, malloc and stdio packages were avoided.
  NOTE: With glibc2, libc-start.o dependencies induce ld to pull in a huge
        fraction of the whole library.  A replacement for the glibc version
	makes null program static binaries only 2..3K.  Initializer, finalizer,
	and atexit() are disabled, but we do not use them here.
*/
#include <sys/types.h>		/* uid_t, gid_t, stuff for linux/string.h */
#ifdef __linux
#   include <linux/string.h>	/* strchr(), strrchr(), strtok() */
#else
#   include <string.h>		/* strchr(), strrchr(), strtok() */
#endif
#include <unistd.h>		/* execv(), stat(), access() */
#include <sys/stat.h>		/* struct stat */
extern void *alloca(unsigned long);

#if defined(__FreeBSD__)
int errno;  /* On BSD static binaries are O(3K) larger just to get errno def */
#endif	    /* since it's in same .o as all the error lists which are big. */

#define ERR(str) write(2, str, strlen(str))

static int recursive = 1;

#include "bname.c"		/* basename locator */
#include "envget.c"		/* optimized getenv with default val */
#include "selfpath.c"
#define N 1024
#include "var_name.c"

/* Instead of using exec*p() we explicitly search $PATH, avoiding a re-exec of
   ourself. This way PATH can remain unmodified for whatever program we wrap. */

static inline
int execv_noloop(char *av0, char ** av) {
    struct stat  st_self, lst_self, st_new;
    char         path[4096],		        /* path buffer */
	        *Path = (char *) 0,	        /* contents of $PATH */
	        *pfx,			        /* pointer into PATH */
	        *eopfx;		         	/* end of dir within PATH */
    int	         nprog = strlen(av[0]),
	         done  = 0;

    if (selfpath(av0, path, &st_self, &lst_self)<0)  /* discover self or die */
	return 1;
    Path = envget("PATH", (char*) 0);		/* self-avoiding exec loop */
    for (pfx = Path; !done && pfx && *pfx; pfx = eopfx) {
	if (av[0][0]=='/' || av[0][0]=='.') {	/* specified path to prog */
	    strncpy(path, av[0], sizeof path);
	    done = 1;
	    goto noloop;
	}
	if ((done = !(eopfx = strchr(pfx, ':'))))
	    eopfx = pfx + strlen(pfx);
	strncpy(path, pfx, eopfx - pfx);
        path[eopfx - pfx] = '/';
	eopfx++;
	strncpy(path + (eopfx - pfx), av[0], nprog + 1);
noloop:
	if (recursive) {			/* avoid only same soft-link */
	    if (lstat(path, &st_new) != 0)
		continue;
	    if (ST_SAME(&st_new, &lst_self))
		continue;
	} else {				/* avoid same hard-link */
	    if (stat(path, &st_new) != 0)
		continue;
	    if (ST_SAME(&st_new, &st_self))
		continue;
	}
	av[0] = path;
	execv(path, av);
    }
    return 1;
}

static inline
void av_append(char **av, int *n, int n_av, char *val) {
    if (*n == 0) {
	int m = strlen(val);

	if (val[m - 1] == '!') {
	    recursive = 0;
	    val[m - 1] = '\0';
	}
    }
    if (++*n < n_av)
	av[*n - 1] = val;
    else
	ERR("alias: too many arguments\n");
}

static inline
void copy_argv(char **av, int *n, int n_av, char **argv, char **xcl, int *nxcl) {
     char **x;
     argv++;
beg: for (; *argv; argv++) {
 	for (x = xcl; *x; x++)
 	    if (strcmp(*x, *argv) == 0) {
 		argv += nxcl[x - xcl];
 		goto beg;
 	    }
 	av_append(av, n, n_av, *argv);
     }
}

int main(int argc, char **argv) {
    int    n = 1,			/* no. elements in new av */
	   n_av = argc + 128,		/* 128 extra alias-y args */
	   did_argv = 0,		/* flag for whether $* used */
	   nneg = 0,			/* number of negative options */
	   nxcl[16];			/* number of excluded args for xcl[] */
#if 0					/* Older C's */
    char **av = (char **)alloca((argc + 128)*sizeof(char *)), /* GNU C-ish */
#else					/* GNU C/ISO C */
    char  *av [argc + 128],
          *xcl[16],
#endif
	  *av0 = argv[0],		 	/* first slot in old argv */
	  *base = bname(av0),			/* last path component */
	  *argse, *arg,				/* args from var, cur arg */
          *IFS  = " \t\n",
           vnm[N] = "ALIAS_",
	   npfx = strlen(vnm);

    if (vnm_init(vnm+npfx, N-npfx, base) < 0)	/* basename -> var name */
	return 1;
    if ((argse = envget(vnm, (char *) 0)))
        --n;					/* so 1st AV_APPEND does 0 */
    else {					/* no new argv[0] */
#if 0
	ERR(av0); ERR(": warning: "); ERR(vnm); ERR(" undefined\n");
#endif
	av[0] = base;	/* XXX base or av0? */
	goto finish;
    }
    if ((arg = strtok(argse, IFS)))		/* build rest of argv */
	do
	    if (arg[0] == '!') {
		for (xcl[nneg] = arg; xcl[nneg][0] == '!'; xcl[nneg]++) /* */;
		nxcl[nneg] = xcl[nneg] - arg;
		nneg++;
	    } else if (strcmp(arg, "$*") == 0) {
		xcl[nneg] = (char *) 0;
		copy_argv(av, &n, n_av, argv, xcl, nxcl);
	        did_argv = 1;
	    } else
		av_append(av, &n, n_av, arg);
	while ((arg = strtok((char *) 0, IFS)));
    else
	av_append(av, &n, n_av, argse);		/* single tok => prog */
finish:
    if (!did_argv) {
	xcl[nneg] = (char *) 0;
	copy_argv(av, &n, n_av, argv, xcl, nxcl);
    }
    av_append(av, &n, n_av, (char *) 0);	/* terminate arg list */
    return execv_noloop(av0, av);		/* try to exec or fail */
}
