#ifndef PSEARCH_C
#define PSEARCH_C

#define ST_SAME(__st1, __st2) ((__st1)->st_dev == (__st2)->st_dev && \
	                       (__st1)->st_ino == (__st2)->st_ino)
/*
  Assuming a standard $PATH search was used and at least the basename of av0 is
  reliable and the $PATH that we inherited to quasi-reliably get the name of
  the link we were executed under.  This necessarily creates a race condition
  (elements of $PATH can change).  That is the least of our concerns since av0
  is not technically reliable anyway.  We allow constraining the path search to
  find a program whose device and inode numbers match those provided in 'st'.
*/
static int path_search(char *prog, struct stat *st, struct stat *lst) {
    char  path[4096],		         /* path buffer */
	 *Path = (char *) 0,	         /* contents of $PATH */
	 *pfx,			         /* pointer into PATH */
	 *eopfx;		         /* end of dir within PATH */
    int	  nprog = strlen(prog),
	  done  = 0;

    Path = envget("PATH", (char*) 0);
    for (pfx = Path; !done && pfx && *pfx; pfx = eopfx) {
	if ((done = !(eopfx = strchr(pfx, ':'))))
	    eopfx = pfx + strlen(pfx);
	strncpy(path, pfx, eopfx - pfx);
        path[eopfx - pfx] = '/';
	eopfx++;
	strncpy(path + (eopfx - pfx), prog, nprog + 1);

	if (stat(path, lst) != 0)
	    continue;
	if (st) {
	    if (ST_SAME(lst, st))
		goto match;
	} else if (access(path, X_OK) == 0)
	    goto match;
    }
    return -1;
match:
    strcpy(prog, path);
    lstat(path, lst);
    return 0;
}
#endif /* PSEARCH_C */
